/*
 * Copyright(c) 2010, Norio Kobota, All rights reserved.
 */

#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <arpa/inet.h>
#include <sys/types.h>
#include <sys/un.h>
#include <sys/socket.h>
#include <netdb.h>
#include <errno.h>
#include <poll.h>

#include "mod_websocket.h"
#include "mod_websocket_socket.h"

int mod_websocket_connect(const char *host, handler_ctx *hctx) {
    int flags, fd = -1, sockret = -1, len = 0;
    struct sockaddr_un address;
    socklen_t socklen = sizeof(sockret);
    struct pollfd pollfd;

    fd = socket(AF_UNIX, SOCK_STREAM, 0);
    if (fd < 0) {
        DEBUG_LOG(MOD_WEBSOCKET_LOG_ERR, "sd", "couldn't create socket for backend: ", errno);
        return fd;
    }
    if ((flags = fcntl(fd, F_GETFL, 0)) < 0 ||
        fcntl(fd, F_SETFL, flags | O_NONBLOCK) < 0) {
        close(fd);
        fd = -1;
        DEBUG_LOG(MOD_WEBSOCKET_LOG_ERR, "s", "fcntl for backend socket failed");
        return fd;
    }
    pollfd.fd = fd;
    pollfd.events = POLLOUT;
    address.sun_family = AF_UNIX;
    strcpy(address.sun_path, host);
    len = sizeof(struct sockaddr_un);
    if (connect(fd, (struct sockaddr*)&address, len) < 0 &&
        errno != EINPROGRESS) {
        close(fd);
        fd = -1;
        DEBUG_LOG(MOD_WEBSOCKET_LOG_ERR, "sd", "socket connect for backend failed: ", errno);
        return fd;
    }
    if (poll(&pollfd, 1, 5000) == 0) {
        close(fd);
        fd = -1;
        DEBUG_LOG(MOD_WEBSOCKET_LOG_ERR, "s", "socket poll for backend failed");
        return fd;
    }
    if ((pollfd.revents & POLLOUT) &&
        getsockopt(fd, SOL_SOCKET, SO_ERROR, &sockret, &socklen) == 0 &&
        sockret == 0 ) {
        return fd;
    }
    close(fd);
    fd = -1;
    DEBUG_LOG(MOD_WEBSOCKET_LOG_ERR, "s", "getsockopt for backend socket failed");
    return fd;
}

void mod_websocket_disconnect(int fd) {
    close(fd);
}

static void store_sockinfo(struct sockaddr_storage *sa, mod_websocket_addrinfo_t *info) {
    assert(sa != NULL && info != NULL);
    if (sa->ss_family == AF_INET) {
        struct sockaddr_in *s = (struct sockaddr_in *)sa;
        info->port = ntohs(s->sin_port);
        inet_ntop(AF_INET, &s->sin_addr, info->addr, sizeof(info->addr));
    } else {
        struct sockaddr_in6 *s = (struct sockaddr_in6 *)sa;
        info->port = ntohs(s->sin6_port);
        inet_ntop(AF_INET6, &s->sin6_addr, info->addr, sizeof(info->addr));
    }
}

int mod_websocket_getsockinfo(int fd, mod_websocket_sockinfo_t *info) {
    socklen_t len;
    struct sockaddr_storage sa;

    if (info == NULL) {
        return -1;
    }
    len = sizeof(sa);
    if (getsockname(fd, (struct sockaddr*)&sa, &len) == -1) {
        return -1;
    }
    store_sockinfo(&sa, &info->self);
    if (getpeername(fd, (struct sockaddr*)&sa, &len) == -1) {
        return -1;
    }
    store_sockinfo(&sa, &info->peer);
    return 0;
}
